from canaryd_packages import six

from canaryd.log import logger


class Change(object):
    def __init__(self, plugin, type_, key, data=None):
        self.plugin = plugin
        self.type = type_
        self.key = key

        # COMPAT w/canaryd < 0.2
        # Legacy support where added/deleted would either have data/None,
        # rather than data changes (where every key: (old_value, new_value)).
        if data and self.type in ('added', 'deleted') and not all(
            isinstance(item, (tuple, list)) and len(item) == 2
            for _, item in six.iteritems(data)
        ):
            if self.type == 'added':
                data = dict(
                    (key, (None, value))
                    for key, value in six.iteritems(data)
                )

            elif self.type == 'deleted':
                data = dict(
                    (key, (value, None))
                    for key, value in six.iteritems(data)
                )

            logger.info('Converted legacy data: {0}'.format(data))

        self.data = data

    def serialise(self):
        return self.type, self.key, self.data

    def __repr__(self):
        return 'Change: {0}'.format((self.plugin, self.type, self.key, self.data))


def make_diffs(plugin, changes):
    return [
        Change(plugin, *change)
        for change in changes
    ]


def get_state_diff(plugin, plugin_state, previous_plugin_state):
    '''
    Diffs two state dicts and returns a list of changes.

    Changes:
        All come as a tuple of ``(plugin, type, key, data=None)``.
    '''

    changes = []

    # Look through the previous state to find any items that have been removed
    for key, previous_item in six.iteritems(previous_plugin_state):
        if key not in plugin_state:
            state_diff = dict(
                (k, (v, None))
                for k, v in six.iteritems(previous_item)
            )

            changes.append(Change(plugin, 'deleted', key, data=state_diff))

    # Loop the new state
    for key, item in six.iteritems(plugin_state):
        previous_item = previous_plugin_state.get(key)

        # Add anything that doesn't exist
        if not previous_item:
            state_diff = dict(
                (k, (None, v))
                for k, v in six.iteritems(item)
            )
            changes.append(Change(plugin, 'added', key, data=state_diff))
            continue

        # Create the diff, which is a key -> (old, new) values
        all_keys = list(six.iterkeys(item)) + list(six.iterkeys(previous_item))
        all_keys = set(all_keys)

        state_diff = dict(
            (k, (previous_item.get(k), item.get(k)))
            for k in all_keys
            if item.get(k) != previous_item.get(k)
        )

        # If something changed - send the event!
        if state_diff:
            # If this plugin disables diffs, remake the state diff w/everything
            if not plugin.diff_updates:
                state_diff = dict(
                    (k, (previous_item.get(k), v))
                    for k, v in six.iteritems(item)
                )

            changes.append(Change(plugin, 'updated', key, data=state_diff))

    return changes



import os
import shlex
import sys

from canaryd_packages import six
# Not ideal but using the vendored in (to requests) chardet package
from canaryd_packages.requests.packages import chardet

from canaryd.log import logger


if os.name == 'posix' and sys.version_info[0] < 3:
    from canaryd_packages.subprocess32 import *  # noqa: F403
else:
    from subprocess import *  # noqa: F403


def ensure_command_tuple(command):
    if not isinstance(command, (list, tuple)):
        return shlex.split(command)
    return command


def decode_output(output):
    if isinstance(output, six.binary_type):
        encoding = chardet.detect(output)['encoding']
        if encoding:
            output = output.decode(encoding)
        else:
            output = output.decode()

    return output


def get_command_output(command, *args, **kwargs):
    logger.debug('Executing command: {0}'.format(command))

    if not kwargs.get('shell', False):
        command = ensure_command_tuple(command)

    try:
        output = check_output(  # noqa: F405
            command,
            close_fds=True,
            stderr=STDOUT,  # noqa: F405
            *args, **kwargs
        )
    except CalledProcessError as e:  # noqa: F405
        e.output = decode_output(e.output)
        raise e

    output = decode_output(output)
    return output



# canaryd
# File: canaryd/canaryctl/__init__.py
# Desc: canaryctl functions

from __future__ import print_function

import json
import logging

from os import path, system
from time import sleep

from canaryd_packages import click, six

from canaryd import remote
from canaryd.exceptions import CanarydError
from canaryd.log import logger, setup_logging
from canaryd.plugin import (
    get_and_prepare_working_plugins,
    get_plugin_by_name,
    get_plugin_states,
    get_plugins,
    prepare_plugin,
)
from canaryd.remote import ApiError, CanaryJSONEncoder, create_event
from canaryd.script import ensure_scripts_directory
from canaryd.settings import (
    ensure_config_directory,
    get_config_directory,
    get_config_file,
    get_settings,
    write_settings_to_config,
)
from canaryd.version import __version__

from .check_root import check_root
from .install_service import install_service


# Parse arguments
@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.option('-v', '--verbose', is_flag=True)
@click.option('-d', '--debug', is_flag=True)
@click.version_option(
    version=__version__,
    prog_name='canaryctl',
    message='%(prog)s: v%(version)s',
)
def main(verbose, debug):
    '''
    canaryd control.
    '''

    log_level = setup_logging(verbose, debug, default_level=logging.WARNING)

    logger.info('Starting canaryctl v{0}'.format(__version__))
    logger.info('Log level set to: {0}'.format(
        logging.getLevelName(log_level),
    ))

    # Ensure the scripts directory (in config) exists
    ensure_config_directory()
    ensure_scripts_directory()


@main.command()
@click.option('--start', is_flag=True, default=False)
@click.option('--enable', is_flag=True, default=False)
@click.argument('key', required=False)
@click.pass_context
def init(ctx, start, enable, key):
    '''
    Create the canaryd service and start it.

    This command will attempt to register if the config file is not found.
    '''

    check_root('Not root user, cannot create services.', exit=True)

    config_file = get_config_file()

    # Register if the config file cannot be found
    if not path.exists(config_file):
        did_register = ctx.invoke(register, key=key)

        if not did_register:
            raise CanarydError(click.style('Failed to register', 'red'))

    # Install the service
    click.echo('--> Installing canaryd service')
    start_command, enable_command = install_service()

    if (
        start or
        click.confirm(
            'Start canaryd service ({0})?'.format(start_command),
            default=True,
        )
    ):
        system(start_command)

    if enable_command and (
        enable or
        click.confirm(
            'Enable canaryd service ({0})?'.format(enable_command),
            default=True,
        )
    ):
        system(enable_command)


@main.command()
@click.argument('key', required=False)
def register(key):
    '''
    Register this server on Service Canary.

    If no api key is provided, you can sign up instantly.
    '''

    check_root('''
Not root user, using config directory:
{0}
It is recommended to run canaryd as root as
plugins that require privileges (eg iptables)
will not function properly.
'''.strip().format(get_config_directory()))

    config_file = get_config_file()

    if not key:
        click.echo('--> No key provided or set for this instance')
        click.echo('--> To sign up, input your email address, or blank to skip')
        email_or_blank = six.moves.input('Email or blank: ')

        if '@' not in email_or_blank:
            return False

        # Signup and get the key
        did_signup, key_or_message = remote.signup(email_or_blank)

        if did_signup:
            key = key_or_message
            click.echo('--> You are now signed up for servicecanary.com.')
            click.echo('--> Check your email for a login link to view updates.')

        else:
            click.echo(click.style(key_or_message, 'blue'))
            return False

    # Register the server
    server_id = remote.register(key=key)

    # Create our settings
    settings = get_settings()
    settings.api_key = key
    settings.server_id = server_id

    # Write the settings to the config file
    write_settings_to_config(settings)

    click.echo('--> {0} written'.format(config_file))

    return True


def _print_plugin_states(states):
    n_states = len(states)

    for i, (plugin, status_data) in enumerate(states, 1):
        status, data = status_data

        click.echo('State for {0}{1}:'.format(
            click.style(plugin.name, bold=True),
            '({0} items)'.format(len(data)) if status else '',
        ))

        if status:
            print(json.dumps(
                data,
                cls=CanaryJSONEncoder,
                indent=4,
            ))
        else:
            raise CanarydError(click.style(
                'Unexpected exception while getting {0} state: {1}({2})'.format(
                    plugin.name,
                    data.__class__.__name__,
                    data,
                ),
                'red',
                bold=True,
            ))

        if i < n_states:
            click.echo()


def _get_all_states(ctx, param, value):
    if not value:
        return

    settings = get_settings()
    plugins = get_and_prepare_working_plugins(settings)

    states = get_plugin_states(plugins, settings)
    _print_plugin_states(states)

    ctx.exit()


@main.command()
@click.option(
    '--all',
    is_flag=True,
    is_eager=True,
    expose_value=False,
    callback=_get_all_states,
)
@click.option(
    '--delay',
    type=int,
    default=0,
    help='Seconds to wait after preparing plugins before collecting state.',
)
@click.argument('plugin')
def state(plugin, delay):
    '''
    Get state for a single plugin.
    '''

    # Trigger load of all plugins first
    get_plugins()

    settings = get_settings()
    target_plugin = get_plugin_by_name(plugin)

    if not target_plugin:
        raise CanarydError(click.style(
            'Invalid plugin: {0}'.format(plugin),
            'red',
        ))

    prepare_status = prepare_plugin(target_plugin, settings)

    if prepare_status is not True:
        _, e = prepare_status
        raise CanarydError(click.style(
            'Plugin unavailable: {0} ({1})'.format(target_plugin.name, e),
            'yellow',
        ))

    sleep(delay)

    states = get_plugin_states([target_plugin], settings)
    _print_plugin_states(states)


@main.command()
def plugins():
    '''
    List available and enabled plugins.
    '''

    click.echo('--> Available plugins: {0}'.format(', '.join(
        plugin.name for plugin in get_plugins()
    )))

    working_plugins = get_and_prepare_working_plugins(get_settings())

    click.echo('--> Enabled plugins:')
    for plugin in working_plugins:
        click.echo(click.style('    {0}'.format(plugin.name), bold=True))


@main.command()
def status():
    '''
    Print the current status of canaryd.
    '''

    config_file = get_config_file()

    click.echo('config file: {0}'.format(config_file))

    if not path.exists(config_file):
        click.echo((
            'No config file ({0}) exists, '
            'please run `canaryctl init`'
        ).format(config_file))
        return

    settings = get_settings()

    try:
        remote.ping(settings)
        status = click.style('online', 'green')

    except ApiError as e:
        e.log()
        status = click.style('offline', 'red')

    click.echo('connection: {0}'.format(status))


@main.command()
@click.argument('plugin')
@click.argument('type')
@click.argument('description')
@click.argument('data', required=False)
def event(plugin, type, description, data=None):
    '''
    Push a custom event to Service Canary.
    '''

    settings = get_settings()

    if data:
        data = json.loads(data)

    response_data = create_event(settings, plugin, type, description, data)

    if response_data['created']:
        click.echo('Event created: {0}'.format(
            click.style(response_data['event_id'], bold=True),
        ))
    else:
        click.echo(click.style('Event not created', 'red'))



from glob import glob
from hashlib import sha1
from os import path, stat

from canaryd_packages import six

from canaryd.plugin import Plugin

# Attempt to import pwd/grp - only supported on *nix
try:
    import pwd
    import grp
except ImportError:
    pwd = None
    grp = None


BUFFER_SIZE = 65536
MAX_SIZE = 1024 ** 3  # 1GB
PATHS_TO_TRACK = (
    '/etc/*',
    '/boot/*',
    '/dev/*',

    '/lib/*',
    '/usr/lib/*',
    '/usr/local/lib/*',
    '/lib64/*',
    '/usr/lib64/*',

    '/bin/*',
    '/sbin/*',
    '/usr/bin/*',
    '/usr/sbin/*',
    '/usr/local/bin/*',
    '/usr/local/sbin/*',
    '/opt/bin/*',
    '/opt/sbin/*',
)
PATHS_TO_IGNORE = (
    '/etc/adjtime',
)


def get_file_hash(filename):
    sha = sha1()

    with open(filename, 'rb') as f:
        while True:
            data = f.read(BUFFER_SIZE)
            if not data:
                break
            sha.update(data)
    return sha.hexdigest()


class Integrity(Plugin):
    '''
    Tracks the integrity of system files by checking their owner, permissions
    and checksum hashes.
    '''

    spec = ('file', {
        'sha1_hash': six.text_type,
        'size': int,
        'user': six.text_type,
        'group': six.text_type,
        'permissions': six.text_type,
    })

    is_slow = True

    def prepare(self, settings):
        return True

    def get_files_to_check(self, settings):
        plugin_settings = settings.get_plugin_settings('integrity')

        paths_to_check = set(PATHS_TO_TRACK)
        user_paths_to_check = plugin_settings.get('check_paths')
        if user_paths_to_check:
            paths_to_check.update(user_paths_to_check.split())

        checked_files = set()

        for path_to_track in paths_to_check:
            for filename in glob(path_to_track):
                if (
                    filename in PATHS_TO_IGNORE
                    or not path.isfile(filename)
                ):
                    continue

                if filename not in checked_files:
                    checked_files.add(filename)
                    yield filename

    def get_state(self, settings):
        file_hashes = {}
        for filename in self.get_files_to_check(settings):
            stat_data = stat(filename)

            if stat_data.st_size > MAX_SIZE:
                continue

            file_data = {
                'size': stat_data.st_size,
                'permissions': format(stat_data.st_mode, 'o'),
            }

            if pwd:
                file_data['user'] = pwd.getpwuid(stat_data.st_uid).pw_name

            if grp:
                file_data['group'] = grp.getgrgid(stat_data.st_gid).gr_name

            try:
                file_data['sha1_hash'] = get_file_hash(filename)
            except (OSError, IOError):
                pass

            file_hashes[filename] = file_data

        return file_hashes



import json

from collections import deque

from canaryd_packages import six

from canaryd.plugin import Plugin

LSHW_CLASSES = ('bridge', 'processor', 'memory', 'disk', 'network')
TOP_LEVEL_DICT_KEYS = ('capabilities', 'configuration')
TOP_LEVEL_STRING_KEYS = (
    'serial', 'version', 'vendor',
    'product', 'handle', 'description',
    'serial_number',  # legacy support
)


def slugify(value):
    return value.lower().replace(' ', '_')


def get_lshw_items(children):
    items = []

    for item in children:
        if 'children' in item:
            items.extend(get_lshw_items(item.pop('children')))

        if item['class'] in LSHW_CLASSES:
            new_item = {
                'type': item.pop('class'),
            }

            for k in TOP_LEVEL_STRING_KEYS + TOP_LEVEL_DICT_KEYS:
                if k in item:
                    new_item[k] = item.pop(k)

            new_item['meta'] = item

            key = item.pop('logicalname', item.pop('id'))

            # Where we have multiple names, use first and attach the others in meta
            if isinstance(key, list):
                new_item['meta']['other_keys'] = key[1:]
                key = key[0]

            items.append((key, new_item))

    return items


def get_hardware_spec():
    hardware_spec = {
        'type': six.text_type,
        'meta': dict,
    }

    for key in TOP_LEVEL_STRING_KEYS:
        hardware_spec[key] = six.text_type

    for key in TOP_LEVEL_DICT_KEYS:
        hardware_spec[key] = dict

    return hardware_spec


class Hardware(Plugin):
    '''
    Tracks system hardware using ``lshw`` (if available).
    '''

    spec = ('key', get_hardware_spec())

    command = 'lshw -json'
    current_state = None

    is_slow = True

    def __init__(self):
        self.previous_states = deque((), 2)

    def parse(self, output):
        '''
        The last two states are tracked and compared such that items change
        after two identical iterations of state.

        This is to account for lshw output changing for one random iteration in
        thousands: https://github.com/Oxygem/canaryd/issues/18
        '''

        # Parse and generate state
        lshw_data = json.loads(output)
        data = get_lshw_items(lshw_data.get('children', []))

        # Rotating list of the latest two states
        self.previous_states.append(dict(data))

        # First call? Just return the state
        if len(self.previous_states) == 1:
            self.current_state = self.previous_states[0]
            return self.current_state

        other_state, latest_state = self.previous_states

        for key, item in six.iteritems(latest_state):
            # New item not in previous state? OK!
            if key not in other_state:
                self.current_state[key] = item

            # Item is same as previous state? OK!
            elif key in other_state and all(
                other_state[key].get(k) == item.get(k)
                for k in TOP_LEVEL_STRING_KEYS
            ):
                self.current_state[key] = item

        # Remove any items dropped from current state in latest_state
        for key, item in list(six.iteritems(self.current_state)):
            if key not in latest_state:
                del self.current_state[key]

        return self.current_state

    @staticmethod
    def should_apply_change(change):
        # If any of our top level keys (serial, vendor, product, etc) change,
        # count as a change. Anything nested (meta, capabilities, etc) is ignored.
        if change.type == 'updated':
            return any(
                key in change.data
                for key in TOP_LEVEL_STRING_KEYS
            )



import re
import socket

from datetime import datetime, timedelta

from canaryd.plugin import Plugin
from canaryd.subprocess import get_command_output

UPTIME_REGEX = re.compile((
    r'up\s+(.*?),\s+[0-9]+ '
    r'users?,\s+load averages?: '
    r'([0-9]+\.[0-9][0-9]),?\s+([0-9]+\.[0-9][0-9]),?\s+([0-9]+\.[0-9][0-9])'
))


def ensure_datetime(datetime_or_string):
    if isinstance(datetime_or_string, datetime):
        return datetime_or_string

    return datetime.strptime(datetime_or_string, '%Y-%m-%dT%H:%M:%S')


def get_uptime():
    data = []
    output = get_command_output('uptime')

    for line in output.splitlines():
        line = line.strip()
        matches = re.search(UPTIME_REGEX, line)

        if matches:
            duration, av1, av5, av15 = matches.groups()

            days = 0
            hours = 0
            mins = 0

            if 'day' in duration:
                match = re.search(r'([0-9]+)\s+day', duration)
                days = int(match.group(1))

            if 'hr' in duration:
                match = re.search(r'([0-9]+)\s+hr', duration)
                hours = int(match.group(1))

            if ':' in duration:
                match = re.search(r'([0-9]+):([0-9]+)', duration)
                hours = int(match.group(1))
                mins = int(match.group(2))

            if 'min' in duration:
                match = re.search(r'([0-9]+)\s+min', duration)
                mins = int(match.group(1))

            up_since = datetime.utcnow() - timedelta(
                days=days, hours=hours, minutes=mins,
            )

            up_since = up_since.replace(
                second=0,
                microsecond=0,
            )

            data.append(('up_since', up_since))

    return data


def get_uname_data(flag):
    output = get_command_output(
        'uname -{0}'.format(flag),
    )

    return output.strip()


class Meta(Plugin):
    '''
    Tracks basic server information (kernel, OS).
    '''

    spec = ('key', {
        'value': None,
    })

    @staticmethod
    def prepare(settings):
        pass

    @staticmethod
    def get_state(settings):
        data = get_uptime()
        data.append(('hostname', socket.gethostname()))
        data.append(('kernel', get_uname_data('s')))
        data.append(('kernel_release', get_uname_data('r')))
        data.append(('arch', get_uname_data('m')))

        # Nest each value in a dict
        return dict(
            (key, {
                'value': value,
            })
            for key, value in data
        )

    @staticmethod
    def should_apply_change(change):
        # Check to see if we rebooted, if not there's no change
        if change.key == 'up_since':
            old_up, new_up = change.data['value']

            if old_up is None:
                return True

            previous_up = ensure_datetime(old_up)
            # Account for slight jitter
            previous_up += timedelta(minutes=1)

            up = ensure_datetime(new_up)

            return previous_up < up
        return True

    @staticmethod
    def get_description_for_change(change):
        if change.type != 'updated':
            return

        if change.key == 'up_since':
            return 'Server rebooted'



# canaryd
# File: canaryd/util.py
# Desc: shared utilities for canaryd, mostly for plugins

import logging
import sys
import traceback

from datetime import datetime
from logging.handlers import SysLogHandler

from canaryd_packages import click
from canaryd_packages import six

STDOUT_LOG_LEVELS = (logging.DEBUG, logging.INFO)
STDERR_LOG_LEVELS = (logging.WARNING, logging.ERROR, logging.CRITICAL)

# Get the logger
logger = logging.getLogger('canaryd')


class LogFilter(logging.Filter):
    def __init__(self, *levels):
        self.levels = levels

    def filter(self, record):
        # Warnings only for vendored packages
        if record.name.startswith('canaryd.packages'):
            return record.levelno >= logging.WARNING

        return record.levelno in self.levels


class LogFormatter(logging.Formatter):
    level_to_format = {
        logging.DEBUG: lambda s: click.style(s, 'green'),
        logging.WARNING: lambda s: click.style(s, 'yellow'),
        logging.ERROR: lambda s: click.style(s, 'red'),
        logging.CRITICAL: lambda s: click.style(s, 'red', bold=True),
    }

    def format(self, record):
        message = record.msg

        if record.args:
            message = record.msg % record.args

        # We only handle strings here
        if isinstance(message, six.string_types):
            if record.levelno in self.level_to_format:
                message = self.level_to_format[record.levelno](message)

            now = datetime.now().replace(microsecond=0).isoformat()

            return '{0} {1} {2}'.format(now, record.levelname, message)

        # If not a string, pass to standard Formatter
        else:
            return super(LogFormatter, self).format(record)


def setup_logging(verbose, debug, default_level=logging.CRITICAL):
    log_level = default_level

    if verbose:
        log_level = logging.INFO

    if debug:
        log_level = logging.DEBUG

    # Set the log level
    logger.setLevel(log_level)

    # Setup a new handler for stdout & stderr
    stdout_handler = logging.StreamHandler(sys.stdout)
    stderr_handler = logging.StreamHandler(sys.stderr)

    # Setup filters to push different levels to different streams
    stdout_filter = LogFilter(*STDOUT_LOG_LEVELS)
    stdout_handler.addFilter(stdout_filter)

    stderr_filter = LogFilter(*STDERR_LOG_LEVELS)
    stderr_handler.addFilter(stderr_filter)

    # Setup a formatter
    formatter = LogFormatter()
    stdout_handler.setFormatter(formatter)
    stderr_handler.setFormatter(formatter)

    # Add the handlers
    logger.addHandler(stdout_handler)
    logger.addHandler(stderr_handler)

    return log_level


def setup_logging_from_settings(settings):
    if settings.debug == 'true':
        logger.setLevel(logging.DEBUG)

    if settings.log_file:
        rotation = settings.log_file_rotation
        count = settings.log_file_rotation_count

        try:
            rotation = int(rotation)
        except (TypeError, ValueError):
            pass

        if isinstance(rotation, int):
            handler = logging.handlers.RotatingFileHandler(
                settings.log_file,
                maxBytes=rotation,
                backupCount=count,
            )

        elif isinstance(rotation, six.string_types):
            handler = logging.handlers.TimedRotatingFileHandler(
                settings.log_file,
                when=rotation,
                backupCount=count,
            )

        else:
            handler = logging.FileHandler(settings.log_file)

        formatter = LogFormatter()
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    if settings.syslog_facility:
        handler = SysLogHandler(facility=settings.syslog_facility)
        logger.addHandler(handler)


def print_exception(debug_only=False):
    if debug_only and logger.level != logging.DEBUG:
        return

    traceback.print_exc()



#!/usr/bin/env python

import sys

from canaryd.subprocess import CalledProcessError, get_command_output

# Run once an hour, smartctl can be intensive
# CANARYD_INTERVAL = 3600

SMART_RETURN_BITS = {
    0: False,  # command line parse error
    1: 'device open failed',
    2: 'SMART command failed',
    3: 'disk failing',
    # We track thresholds via stdout
    4: True,  # pre-fail attrs <= thresh
    5: True,  # pre-fail attrs <= thresh in past
    # We track errors via stdout
    6: True,  # error log
    7: True,  # self-test errors
}


try:
    # Ensure smartctl is present & working
    get_command_output(
        'smartctl --version',
    )

except (CalledProcessError, OSError):
    print('No smartctl found!')
    sys.exit(1)


try:
    # Look for OSX disks first (OSX has /dev/sdt which breaks below)
    disks_data = get_command_output(
        'ls /dev/disk?',
        shell=True,
    )

except (CalledProcessError, OSError):
    try:
        # Look for Linux disks (/dev/sdX, /dev/hdX, /dev/nvmeX)
        disks_data = get_command_output(
            'ls /dev/[hs]d? || ls /dev/nvme?',
            shell=True,
        )
    except (CalledProcessError, OSError):
        print('No disks to check found!')
        sys.exit(1)


# Turn into a list of disks
disks = disks_data.strip().split()

warnings = []
criticals = []

for disk in disks:
    try:
        # List health and attributes
        smart_data = get_command_output(
            'smartctl -a {0}'.format(disk),
        )

    except CalledProcessError as e:
        smart_data = e.output

        # If the device doesn't support SMART not much we can do about it!
        if 'Operation not supported by device' in smart_data:
            continue

        for i in SMART_RETURN_BITS:
            bit = e.returncode & i
            if bit and i in SMART_RETURN_BITS:
                return_bit = SMART_RETURN_BITS[i]

                if return_bit is False:
                    raise

                if return_bit is True:
                    continue

                criticals.append('Disk {0} error: {1}'.format(disk, return_bit))
                break

    for line in smart_data.splitlines():
        # First check SMART's overall status
        if line.startswith('SMART overall-health self-assessment test result'):
            bits = line.split()
            test_result = bits[-1]

            if test_result != 'PASSED':
                criticals.append((
                    'Disk {0} has failed test, status: {1}'
                ).format(disk, test_result))
                continue

        # Ignore any non-attribute line
        if all(type_ not in line for type_ in ('Pre-fail', 'Old_age')):
            continue

        bits = line.split()
        name = bits[1]
        value = int(bits[3])
        threshold = int(bits[5])

        # Skip where both the value and threshold are 0, like power_on_hours/etc
        if value == 0 and threshold == 0:
            continue

        # From: https://superuser.com/a/354254
        # Each Attribute also has a Threshold value (whose range is 0 to 255)
        # which is printed under the heading "THRESH". If the Normalized value
        # is less than or equal to the Threshold value, then the Attribute is
        # said to have failed. If the Attribute is a pre-failure Attribute,
        # then disk failure is imminent.
        if value == threshold:
            warnings.append((
                'Attrbute {0} on disk {1} is at the threshold ({2})'
            ).format(name, disk, threshold))

        elif value < threshold:
            criticals.append((
                'Attribute {0} on disk {1} is below the threshold ({2}/{3})'
            ).format(name, disk, value, threshold))


exit_code = 0

if warnings:
    exit_code = 1
    print('\n'.join(warnings))

if criticals:
    exit_code = 2
    print('\n'.join(criticals))

if exit_code == 0:
    print('All disks healthy')


sys.exit(exit_code)



from __future__ import division

import re
import sys

from multiprocessing import cpu_count
from os import path, sep as os_sep

from canaryd_packages import six

from canaryd.subprocess import get_command_output


def get_ps_cpu_stats():
    '''
    Uses ps + awk to total CPU usage, then divide by # CPUs to get the final %.
    '''

    output = get_command_output(
        "ps -A -o %cpu | awk '{s+=$1} END {print s}'",
        shell=True,
    )

    try:
        cpu_percentage = float(output.strip())

    except (TypeError, ValueError):
        return {}

    return {
        'cpu': {
            'percentage': cpu_percentage / cpu_count(),
        },
    }


def get_proc_cpu_stats():
    '''
    Getsa + parses 2x `/proc/stat` output to calculate CPU %.
    '''

    output = get_command_output(
        'cat /proc/stat && sleep 1 && cat /proc/stat',
        shell=True,
    )

    # Parse /proc/stat
    columns = ('user', 'nice', 'system', 'idle', 'iowait', 'irq', 'soft_irq')
    cpus = {}
    diffs = None
    total = None

    for line in output.splitlines():
        if not line:
            continue

        bits = line.split()
        key, details = bits[0], [int(bit) for bit in bits[1:8]]

        if key != 'cpu':
            continue

        if key in cpus:
            # Calculate usage (second cat)
            diffs = dict(
                (column, details[i] - cpus[key][i])
                for i, column in enumerate(columns)
            )
            total = sum(six.itervalues(diffs))

        else:
            # Set details (first cat)
            cpus[key] = details

    idle_and_iowait = diffs['idle'] + diffs['iowait']

    return {
        'cpu': {
            'current_value': total - idle_and_iowait,
            'current_max': total,
            'percentage': (total - idle_and_iowait) / total * 100,
        },
        'iowait': {
            'current_value': diffs['iowait'],
            'current_max': total,
            'percentage': diffs['iowait'] / total * 100
        },
    }


def get_cpu_stats():
    # If `ps` command used instead of preferred /proc/stat, quickly work
    # out the CPU % and use that (no iowait support).
    if not path.exists('/proc'):
        return get_ps_cpu_stats()

    # Parse /proc/stat
    return get_proc_cpu_stats()


def parse_memory_stats(
    lines,
    value_is_last=False,
    value_multiplier=1,
    value_divisor=1,
):
    key_map = {
        'MemTotal:': 'memory_total',
        'MemFree:': 'memory_free',
        'Buffers:': 'memory_free',
        'Cached:': 'memory_free',
        'SReclaimable:': 'memory_free',
        'Shmem:': 'memory_free',

        'SwapTotal:': 'swap_total',
        'SwapFree:': 'swap_free',
        'SwapCached:': 'swap_free',

        'Pages free:': 'memory_free',
        'Pages inactive:': 'memory_free',
    }

    stats = {}

    for line in lines:
        bits = line.split()

        if value_is_last:
            key = ' '.join(bits[:-1])
            value = bits[-1][:-1]
        else:
            key, value = bits[:2]

        if key in key_map:
            value = int(value) * value_multiplier / value_divisor

            if key_map[key] in stats:
                stats[key_map[key]] += value
                continue

            stats[key_map[key]] = value

    return stats


def get_proc_memory_stats():
    '''
    Gets + parses `/proc/meminfo` output.
    '''

    output = get_command_output(
        'cat /proc/meminfo',
    )

    stats = parse_memory_stats(
        output.splitlines(),
        value_divisor=1024,
    )

    data = {
        'memory': {
            'current_value': int(stats['memory_total'] - stats['memory_free']),
            'current_max': int(stats['memory_total']),
            'percentage': round(
                (stats['memory_total'] - stats['memory_free']) / stats['memory_total'] * 100,
                2,
            ),
        },
    }

    if stats['swap_total'] > 0:
        data['swap'] = {
            'current_value': int(stats['swap_total'] - stats['swap_free']),
            'current_max': int(stats['swap_total']),
            'percentage': round(
                (stats['swap_total'] - stats['swap_free']) / stats['swap_total'] * 100,
                2,
            ),
        }

    return data


def get_vm_stat_memory_stats():
    output = get_command_output(
        'sysctl hw.memsize',
    )

    bits = output.split()
    total_memory = int(bits[1]) / 1024 / 1024

    output = get_command_output(
        'vm_stat',
    )

    stats = parse_memory_stats(
        output.splitlines(),
        value_is_last=True,
        value_multiplier=4096,
        value_divisor=1024 * 1024,
    )

    return {
        'memory': {
            'current_value': int(total_memory - stats['memory_free']),
            'current_max': int(total_memory),
            'percentage': round(
                (total_memory - stats['memory_free']) / total_memory * 100,
                2,
            ),
        },
    }


def get_memory_stats():
    # If `ps` command used instead of preferred /proc/stat, quickly work
    # out the CPU % and use that (no iowait support).
    if sys.platform == 'darwin':
        return get_vm_stat_memory_stats()

    # Parse /proc/stat
    return get_proc_memory_stats()


def get_disk_stats():
    '''
    Parses df output.
    '''

    output = get_command_output('df -kP')

    devices = {}

    for line in output.splitlines():
        bits = line.split()
        filesystem = bits[0]

        if not filesystem.startswith('/'):
            continue

        value = int(bits[2])

        available = int(bits[3])
        max_ = int(bits[1])

        percentage_free = round(available / max_ * 100, 2)
        percentage = 100 - percentage_free

        # Loop backwards through the line bits until we find the bit starting
        # with "/". This means we capture the entire mount path, even if it
        # includes spaces, but also prevents us capturing too much (since df
        # output varies between platforms, like macOS).
        mount_bits = []
        # We know the first few are always blocks/size/used/etc
        possible_bits = bits[2:]

        while possible_bits:
            latest_bit = possible_bits.pop()
            mount_bits.insert(0, latest_bit)

            if latest_bit.startswith('/'):
                break

        mount = ' '.join(mount_bits)

        devices[mount] = {
            'percentage': percentage,
            'current_value': int(round(value / 1024)),
            'current_max': int(round(max_ / 1024)),
        }

    return devices


def _get_device_speed(device):
    speed = 0
    speed_file = path.join(os_sep, 'sys', 'class', 'net', device, 'speed')

    if not path.exists(speed_file):
        return

    try:
        with open(speed_file, 'r') as f:
            data = f.read()
    except IOError:
        return

    speed = int(data)

    # Turn mbits -> bytes
    speed = speed * 125000
    return speed


def get_proc_network_stats():
    '''
    Gets + parses 2x `/proc/net/dev` output.
    '''

    output = get_command_output(
        'cat /proc/net/dev && sleep 1 && cat /proc/net/dev',
        shell=True,
    )

    match_device = r'^\s*([a-zA-Z0-9]+):\s+'

    device_stats = {}
    devices = {}

    for line in output.splitlines():
        matches = re.match(match_device, line)
        if not matches:
            continue

        key, bits = matches.group(1), line.split()

        if key == 'lo':
            continue

        in_bytes = int(bits[1])
        out_bytes = int(bits[9])

        if key in device_stats:
            # Calculate speed + percentage if possible (second cat)
            speed = _get_device_speed(key)

            receive_bytes = in_bytes - device_stats[key][0]
            transmit_bytes = out_bytes - device_stats[key][1]

            receive_percentage = 0
            transmit_percentage = 0
            if speed:
                receive_percentage = receive_bytes / speed * 100
                transmit_percentage = transmit_bytes / speed * 100

            devices['{0}_in'.format(key)] = {
                'current_value': receive_bytes,
                'current_max': speed,
                'percentage': receive_percentage,
            }

            devices['{0}_out'.format(key)] = {
                'current_value': transmit_bytes,
                'current_max': speed,
                'percentage': transmit_percentage,
            }
        else:
            # Assign in/out (first cat)
            device_stats[key] = (in_bytes, out_bytes)

    return devices


def get_netstat_network_stats():
    output = get_command_output(
        'netstat -ib && sleep 1 && netstat -ib',
        shell=True,
    )

    lines = output.splitlines()
    headers = lines[0]
    header_to_position = dict(
        (header, i)
        for i, header in enumerate(headers.split())
    )

    # Collect the metrics from each loop
    first_device_stats = {}
    second_device_stats = {}
    loop_target = first_device_stats

    for line in lines[1:]:
        # Move onto second loop - we store separately as interfaces might be
        # duplicated in the output, and we only want to count their bytes once.
        if line == headers:
            loop_target = second_device_stats
            continue

        bits = line.split()
        key = bits[header_to_position['Name']]

        if key == 'lo0':
            continue

        network = bits[header_to_position['Network']]
        if '<Link' in network:
            continue

        in_bytes = int(bits[header_to_position['Ibytes']])
        out_bytes = int(bits[header_to_position['Obytes']])

        loop_target[key] = (in_bytes, out_bytes)

    # Now loop back and calculate speeds over the 1s
    devices = {}

    for key, stats in second_device_stats.items():
        in_bytes, out_bytes = stats

        receive_bytes = in_bytes - first_device_stats[key][0]
        transmit_bytes = out_bytes - first_device_stats[key][1]

        devices['{0}_in'.format(key)] = {
            'current_value': receive_bytes,
            'percentage': 0.0,
        }

        devices['{0}_out'.format(key)] = {
            'current_value': transmit_bytes,
            'percentage': 0.0,
        }

    return devices


def get_network_stats():
    if path.exists('/proc'):
        return get_proc_network_stats()
    return get_netstat_network_stats()



from canaryd_packages.click import ClickException


class CanarydError(ClickException):
    '''
    Generic canaryd exception.
    '''

    def __init__(self, message='exiting!', *args, **kwargs):
        return super(CanarydError, self).__init__(message, *args, **kwargs)


class UserCancelError(CanarydError):
    '''
    Triggered when a user cancels an action.
    '''

    def __init__(self):
        return super(UserCancelError, self).__init__('User cancelled')


class ConfigError(CanarydError):
    '''
    Triggered when the config file is invalid/broken/missing.
    '''



from distutils.spawn import find_executable
from os import access, listdir, makedirs, path, remove, symlink, X_OK
from shutil import copy

from canaryd.log import logger
from canaryd.settings import get_scripts_directory

INVALID_EXTENSIONS = (
    '.pyc',
)

SCRIPT_SETTING_KEYS = {
    'INTERVAL': int,
}


class ScriptException(Exception):
    pass


class NoScriptFoundError(ScriptException):
    '''
    Raised when a script file cannot be located.
    '''


class NoScriptChangesError(ScriptException):
    '''
    Raised when no changes are made when enabling/disabling scripts.
    '''


class ScriptNotLinkError(ScriptException):
    '''
    Raised when the script is not a link when disabling scripts.
    '''


def _extract_script_settings(full_path):
    settings = {}

    with open(full_path) as f:
        for line in f.readlines():
            for key, parser in SCRIPT_SETTING_KEYS.items():
                match_key = '# CANARYD_{0}'.format(key)
                if not line.startswith(match_key) or '=' not in line:
                    continue

                _, value = line.split('=', 1)
                value = value.strip()
                value = parser(value)

                if value:
                    settings[key] = value

    return settings


def _get_scripts(dirname):
    scripts = []
    all_settings = {}

    script_names = listdir(dirname)
    for name in script_names:
        if any(name.endswith(e) for e in INVALID_EXTENSIONS):
            continue

        full_path = path.join(dirname, name)
        if not access(full_path, X_OK):
            continue

        scripts.append(name)
        settings = _extract_script_settings(full_path)
        if settings:
            all_settings[name] = settings

    return scripts, all_settings


def get_scripts(settings):
    script_settings = {}

    available_scripts, available_script_settings = _get_scripts(
        path.join(get_scripts_directory(), 'available'),
    )
    script_settings.update(available_script_settings)

    enabled_scripts, enabled_script_settings = _get_scripts(
        path.join(get_scripts_directory(), 'enabled'),
    )
    script_settings.update(enabled_script_settings)

    all_scripts = set(available_scripts + enabled_scripts)

    return (
        (script, script in enabled_scripts, script_settings.get(script))
        for script in all_scripts
    )


def enable_script(script, raise_if_noop=True):
    scripts_directory = get_scripts_directory()

    source_script = path.join(scripts_directory, 'available', script)
    link_name = path.join(scripts_directory, 'enabled', script)

    if not path.exists(source_script):
        raise NoScriptFoundError(source_script)

    if path.exists(link_name):
        if not raise_if_noop:
            return
        raise NoScriptChangesError()

    symlink(
        source_script,
        link_name,
    )


def disable_script(script, raise_if_noop=True):
    scripts_directory = get_scripts_directory()

    source_script = path.join(scripts_directory, 'available', script)
    link_name = path.join(scripts_directory, 'enabled', script)

    if not path.exists(source_script):
        raise NoScriptFoundError(source_script)

    if not path.exists(link_name):
        if not raise_if_noop:
            return
        raise NoScriptChangesError()

    if not path.islink(link_name):
        raise ScriptNotLinkError(
            'Script {0} is not a link. You should move it to: {1}.'.format(
                link_name, source_script,
            ),
        )

    remove(link_name)


def ensure_scripts_directory():
    # Make sure the scripts directory exists
    scripts_directory = get_scripts_directory()

    if not path.exists(scripts_directory):
        logger.debug('Creating scripts directory: {0}'.format(scripts_directory))

        # Make the scripts, scripts/enabled & scripts/available directories
        makedirs(path.join(scripts_directory, 'enabled'))
        makedirs(path.join(scripts_directory, 'available'))

        copy_builtin_scripts()


def copy_builtin_scripts(enable_where_possible=True):
    # Copy built in scripts to the scripts/available directory
    available_scripts_directory = path.join(get_scripts_directory(), 'available')

    logger.debug(
        'Copying default scripts to: {0}'.format(available_scripts_directory),
    )

    builtin_scripts_directory = path.join(path.dirname(__file__), 'scripts')

    copied = []
    enabled = []

    for file in listdir(builtin_scripts_directory):
        copy(
            path.join(builtin_scripts_directory, file),
            path.join(available_scripts_directory, file),
        )
        copied.append(file)

    if enable_where_possible:
        script_to_check = {
            'raid_health.py': path.exists('/proc/mdstat'),
            'disk_health.py': find_executable('smartctl'),
        }

        for script, can_enable in script_to_check.items():
            if can_enable:
                enable_script(script, raise_if_noop=False)
                enabled.append(script)

    return copied, enabled



# canaryd
# File: setup.py
# Desc: canaryd package setup

'''
Due to the low-level nature of canaryd, this setup.py targets both setuptools
for systems that have it and the stdlib distutils for those that don't.
'''

SCRIPTS = (
    'bin/canaryctl',
    'bin/canaryd',
)

PACKAGES = (
    'canaryd',
    'canaryd.ctl',
    'canaryd.plugins',

    # Packages
    'canaryd_packages',
    'canaryd_packages.click',

    # Requests package
    'canaryd_packages.requests',
    'canaryd_packages.requests.packages',
    'canaryd_packages.requests.packages.chardet',
    'canaryd_packages.requests.packages.urllib3',
    'canaryd_packages.requests.packages.urllib3.contrib',
    'canaryd_packages.requests.packages.urllib3.packages',
    'canaryd_packages.requests.packages.urllib3.packages.backports',
    'canaryd_packages.requests.packages.urllib3.packages.ssl_match_hostname',
    'canaryd_packages.requests.packages.urllib3.util',
)

TEST_REQUIRES = (
    'nose==1.3.7',
    'jsontest==1.3',
    'coverage==4.5.1',
    'mock==1.3.0',
    'dictdiffer==0.6.1',
)

DEV_REQUIRES = TEST_REQUIRES + (
    # Releasing
    'wheel',
    'twine==1.11.0',

    # Dev debugging
    'ipdb',
    'ipdbplugin',
)


version_data = {}
with open('canaryd/version.py') as f:
    exec(f.read(), version_data)


# distutils/setuptools compatbile kwargs
setup_kwargs = {
    'name': 'canaryd',
    'version': version_data['__version__'],
    'description': 'Client for Service Canary',
    'author': 'Oxygem',
    'author_email': 'hello@oxygem.com',
    'license': 'MIT',
    'url': 'https://servicecanary.com',
    'packages': PACKAGES,
    'scripts': SCRIPTS,
    'include_package_data': True,
    # This must match the contents of MANIFEST.in, to provide full support for
    # distutil, setuptools everywhere. Annoying.
    'package_data': {
        'canaryd': [
            'init_scripts/*',
            'scripts/*',
        ],
        'canaryd_packages.requests': [
            '*.pem',
        ],
    },
}


# Attatch extras if setuptools is present
try:
    from setuptools import setup

    setup_kwargs['extras_require'] = {
        'dev': DEV_REQUIRES,
        'test': TEST_REQUIRES,
    }

# Otherwsie fallback to distutils
except ImportError:
    from distutils.core import setup

setup(**setup_kwargs)



from canaryd_packages import click

from canaryd.exceptions import CanarydError
from canaryd.script import (
    copy_builtin_scripts,
    disable_script,
    enable_script,
    get_scripts,
    NoScriptChangesError,
    NoScriptFoundError,
    ScriptNotLinkError,
)
from canaryd.settings import (
    get_settings,
)

from . import main


@main.group(invoke_without_command=True)
@click.pass_context
def scripts(ctx):
    '''
    List and manage scripts for canaryd.

    \b
    # List scripts
    canaryctl scripts

    \b
    # Enable a script
    canaryctl scripts enable <script.sh>

    \b
    # Disable a script
    canaryctl scripts disable <script.sh>
    '''

    if ctx.invoked_subcommand is not None:
        return

    scripts = get_scripts(get_settings())

    click.echo('--> Scripts:')

    for name, enabled, settings in scripts:
        click.echo('    {0}, enabled: {1}, settings: {2}'.format(
            click.style(name, bold=True),
            enabled,
            settings,
        ))


@scripts.command()
@click.option('--no-enable', is_flag=True, default=False)
def install(no_enable=False):
    '''
    Install the builtin scripts and enable where possible.
    '''

    copied, enabled = copy_builtin_scripts(enable_where_possible=not no_enable)
    click.echo('--> Installed {0}, enabled {1}'.format(
        ', '.join(click.style(s, bold=True) for s in copied),
        ', '.join(click.style(s, bold=True) for s in enabled),
    ))


@scripts.command()
@click.argument('script')
def enable(script):
    '''
    Enable a script.
    '''

    try:
        enable_script(script)

    except NoScriptFoundError as e:
        raise CanarydError(click.style(
            'No script file ({0}) exists.'.format(e.message),
            'red',
        ))

    except NoScriptChangesError:
        click.echo('Script {0} is already enabled.'.format(
            click.style(script, bold=True),
        ))
        return

    click.echo('Script enabled: {0}'.format(click.style(script, bold=True)))


@scripts.command()
@click.argument('script')
def disable(script):
    '''
    Disable a script.
    '''

    try:
        disable_script(script)

    except NoScriptFoundError as e:
        raise CanarydError(click.style(
            'No script file ({0}) exists.'.format(e.message),
            'red',
        ))

    except ScriptNotLinkError as e:
        raise CanarydError(click.style(e.message, 'yellow'))

    except NoScriptChangesError:
        click.echo('Script {0} is already disabled.'.format(
            click.style(script, bold=True),
        ))
        return

    click.echo('Script disabled: {0}'.format(click.style(script, bold=True)))



# canaryd
# File: canaryd/ctl/__main__.py
# Desc: entry point for canaryctl

from __future__ import print_function

from canaryd.remote import ApiError

from . import main, scripts  # noqa


try:
    main()

except ApiError as e:
    e.log()

except Exception:
    # TODO: public Sentry logging

    raise



from canaryd_packages import six

from canaryd.log import logger
from canaryd.plugin import Plugin
from canaryd.subprocess import CalledProcessError, get_command_output

# Attempt to import pwd/grp - the users plugin is *nix only
try:
    import pwd
    import grp
except ImportError:
    pwd = None
    grp = None


def get_last_login_times(timeout):
    output = get_command_output(
        'lastlog',
        timeout=timeout,
    )

    user_logins = {}

    for line in output.splitlines()[1:]:
        user, _, ip, time = line.split(None, 3)
        if time.endswith('*'):
            continue

        user_logins[user] = {
            'login_time': time,
            'login_ip': ip,
        }

    return user_logins


class Users(Plugin):
    '''
    Tracks system users and groups, as well as login events.
    '''

    spec = ('user', {
        'group': six.text_type,
        'groups': [six.text_type],
        'home': six.text_type,
        'shell': six.text_type,
        'login_time': six.text_type,
        'login_ip': six.text_type,
    })

    def prepare(self, settings):
        if any(var is None for var in (pwd, grp)):
            raise self.PrepareError('Missing either pwd or group modules.')

    def get_state(self, settings):
        timeout = self.get_timeout(settings)

        try:
            user_logins = get_last_login_times(timeout)
        except (CalledProcessError, OSError) as e:
            logger.warning('Could not get last login times: {0}'.format(e))
            user_logins = {}

        users = {}

        # Get all groups and map by ID -> group
        groups_data = grp.getgrall()

        groups_by_id = dict(
            (group.gr_gid, group)
            for group in groups_data
        )

        # Get basic user data
        users_data = pwd.getpwall()

        for user in users_data:
            data = {
                'groups': set(),
                'home': user.pw_dir,
                'shell': user.pw_shell,
            }

            username = user.pw_name

            if username in user_logins:
                data.update(user_logins[username])

            users[username] = data

            group = groups_by_id.get(user.pw_gid)
            if group:
                users[user.pw_name]['group'] = group.gr_name

        # Loop back through groups and apply any additional to users
        for group in groups_data:
            for username in group.gr_mem:
                if username not in users:
                    continue

                users[username]['groups'].add(group.gr_name)

        for user_data in six.itervalues(users):
            if 'groups' in user_data:
                user_data['groups'] = sorted(list(user_data['groups']))

        return users

    @staticmethod
    def get_action_for_change(change):
        if change.type != 'updated':
            return

        if 'login_time' in change.data:
            return 'logged in'



import re

from collections import defaultdict
from os import listdir, path, sep as os_sep

from canaryd.log import logger
from canaryd.subprocess import CalledProcessError, get_command_output

# We ignore these as they regularly get deleted/added as part of normal OSX
# lifecycle - and as such any events generated are not of use.
LAUNCHCTL_IGNORE_NAMES = ('oneshot', 'mdworker', 'mbfloagent')

# Systemd service types to ignore
SYSTEMD_IGNORE_TYPES = ('oneshot',)
SYSTEMD_REGEX = re.compile(r'^([a-z\-]+)\.service\s+[a-z\-]+\s+[a-z]+\s+([a-z]+)')

UPSTART_REGEX = re.compile(r'^([a-z\-]+) [a-z]+\/([a-z]+),?\s?(process)?\s?([0-9]+)?')
SUPERVISOR_REGEX = re.compile(r'([a-z\-]+)\s+([A-Z]+)\s+pid\s([0-9]+)')

INITD_USAGE_REGEX = re.compile(r'Usage:[^\n]+status')
INITD_STATUS_REGEX = re.compile(r'\(pid\s+([0-9]+)\)')
# Known init scripts that either don't support status or don't handle it well/quickly
IGNORE_INIT_SCRIPTS = [
    'networking', 'udev-post', 'halt', 'killall',
    'kcare', 'vz',
]

# We *require* procfs to check PID -> port mappings
HAS_PROCFS = path.exists('/proc')
PID_TO_PORTS = defaultdict(set)


def _get_lsof_pid_to_listens(timeout):
    output = get_command_output(
        'lsof -i -n -P -b -l -L -s TCP:LISTEN',
        timeout=timeout,
    )

    for line in output.splitlines():
        # Skip bad/error lines
        if 'no pwd entry' in line:
            continue

        try:
            _, pid, _, _, ip_type, _, _, _, ip_host, _ = line.split(None, 9)

            ip_type = ip_type.lower()
            pid = int(pid)

            # Work out the host:port bit
            host, port = ip_host.rsplit(':', 1)
            port = int(port)

            host_port_tuple = (ip_type, host, port)
            PID_TO_PORTS[pid].add(host_port_tuple)
        except ValueError:
            logger.warning('Dodgy lsof line ignored: "{0}"'.format(line))


def _get_netstat_pid_to_listens(timeout):
    output = get_command_output(
        'netstat -plnt',
        timeout=timeout,
    )

    lines = output.splitlines()

    for line in lines[2:]:
        try:
            bits = line.split()
            proto, _, _, local_address, _, _, program = bits

            # Get the pid from PID/PROGRAM
            pid = program.split('/')[0]
            pid = int(pid)

            # Work out the host:port bit
            host, port = local_address.rsplit(':', 1)
            port = int(port)

            ip_type = 'ipv6' if proto == 'tcp6' else 'ipv4'

            host_port_tuple = (ip_type, host, port)
            PID_TO_PORTS[pid].add(host_port_tuple)
        except ValueError:
            logger.warning('Dodgy netstat line ignored: "{0}"'.format(line))


def get_pid_to_listens(timeout):
    if not HAS_PROCFS:
        return PID_TO_PORTS

    # Loop through the results and cleanup any PIDs that don't exist
    for pid in list(PID_TO_PORTS.keys()):
        if not path.exists('/proc/{0}'.format(pid)):
            PID_TO_PORTS.pop(pid)

    try:
        _get_lsof_pid_to_listens(timeout=timeout)
    except (CalledProcessError, OSError):
        logger.warning('Missing lsof, defaulting to netstat')

        try:
            _get_netstat_pid_to_listens(timeout=timeout)
        except (CalledProcessError, OSError):
            pass

    return PID_TO_PORTS


def get_launchd_services(timeout):
    '''
    Execute & parse ``launchctl list``.
    '''

    output = get_command_output(
        'launchctl list',
        timeout=timeout,
    )

    services = {}

    for line in output.splitlines():
        if any(name in line for name in LAUNCHCTL_IGNORE_NAMES):
            continue

        bits = line.split()

        if not bits or bits[0] == 'PID':
            continue

        name = bits[2]

        # If the last "bit" is just a number, it's the PID, so we strip it
        name_bits = name.split('.')
        if name_bits[-1].isdigit():
            name = '.'.join(name_bits[:-1])

        data = {
            'init_system': 'launchd',
        }

        try:
            data['pid'] = int(bits[0])
        except ValueError:
            pass

        data['running'] = 'pid' in data

        services[name] = data

    return services


def get_initd_services(existing_services, timeout):
    existing_services = existing_services or []

    init_dir = path.join(os_sep, 'etc', 'init.d')
    service_names = listdir(init_dir)
    services = {}

    for name in service_names:
        if name in existing_services or name in IGNORE_INIT_SCRIPTS:
            continue

        script_path = path.join(init_dir, name)

        with open(script_path) as script:
            script_data = script.read()

            if (
                # LSB header
                '### BEGIN INIT INFO' not in script_data
                or not INITD_USAGE_REGEX.search(script_data)
            ):
                IGNORE_INIT_SCRIPTS.append(name)
                continue

        # Get the status
        status = False
        pid = None

        try:
            status_line = get_command_output(
                (script_path, 'status'),
                timeout=timeout,
            )
            status = True

            matches = INITD_STATUS_REGEX.search(status_line)
            if matches:
                pid = int(matches.group(1))

        except (OSError, CalledProcessError):
            pass

        # Check if enabled
        enabled = False

        try:
            found_links = get_command_output(
                'find /etc/rc*.d/S*{0} -type l'.format(name),
                timeout=timeout,
            )

            if found_links.strip():
                enabled = True

        except CalledProcessError:
            pass

        services[name] = {
            'running': status or isinstance(pid, int),
            'pid': pid,
            'enabled': enabled,
            'init_system': 'initd',
        }

    return services


def get_systemd_services(timeout):
    output = get_command_output(
        'systemctl -alt service list-units',
        timeout=timeout,
    )

    services = {}

    for line in output.splitlines():
        matches = SYSTEMD_REGEX.match(line.strip())
        if not matches:
            continue

        name = matches.group(1)

        # Get service info to extract pid/enabled status
        service_output = get_command_output(
            'systemctl show {0}.service'.format(name),
        )

        service_meta = {}
        for line in service_output.splitlines():
            key, value = line.split('=', 1)
            service_meta[key] = value

        if service_meta.get('Type') in SYSTEMD_IGNORE_TYPES:
            continue

        pid = service_meta.get(
            'ExecMainPID',
            service_meta.get('MainPID', None),
        )

        if pid:
            pid = int(pid)

        enabled = service_meta.get('UnitFileState') == 'enabled'

        services[name] = {
            'running': matches.group(2) == 'running',
            'pid': pid,
            'enabled': enabled,
            'init_system': 'systemd',
        }

    return services


def get_upstart_services(timeout):
    output = get_command_output(
        'initctl list',
        timeout=timeout,
    )

    services = {}

    for line in output.splitlines():
        matches = UPSTART_REGEX.match(line)
        if not matches:
            continue

        name = matches.group(1)
        pid = matches.group(4)

        if pid:
            pid = int(pid)

        enabled = True

        # Check if enabled by looking in any override file
        override_filename = path.join(
            os_sep, 'etc', 'init',
            '{0}.override'.format(name),
        )
        if path.exists(override_filename):
            with open(override_filename) as script:
                if 'manual' in script.read():
                    enabled = False

        services[name] = {
            'running': matches.group(2) == 'running',
            'pid': pid,
            'enabled': enabled,
            'init_system': 'upstart',
        }

    return services


def get_supervisor_services(timeout):
    output = get_command_output(
        'supervisorctl status',
        timeout=timeout,
    )

    services = {}

    for line in output.splitlines():
        matches = SUPERVISOR_REGEX.match(line)
        if not matches:
            continue

        name = matches.group(1)
        status = matches.group(2)
        pid = matches.group(3)

        services[name] = {
            'running': status == 'RUNNING',
            'pid': int(pid),
            'init_system': 'supervisor',
        }

    return services
